#ifndef TGTADM_ERROR_H
#define TGTADM_ERROR_H

enum tgtadm_errno {
	TGTADM_SUCCESS,
	TGTADM_UNKNOWN_ERR,
	TGTADM_NOMEM,
	TGTADM_NO_DRIVER,
	TGTADM_NO_TARGET,

	TGTADM_NO_LUN,
	TGTADM_NO_SESSION,
	TGTADM_NO_CONNECTION,
	TGTADM_NO_BINDING,
	TGTADM_TARGET_EXIST,
	TGTADM_LUN_EXIST,
	TGTADM_BINDING_EXIST,

	TGTADM_ACL_EXIST,
	TGTADM_ACL_NOEXIST,
	TGTADM_USER_EXIST,
	TGTADM_NO_USER,
	TGTADM_TOO_MANY_USER,
	TGTADM_INVALID_REQUEST,

	TGTADM_OUTACCOUNT_EXIST,
	TGTADM_TARGET_ACTIVE,
	TGTADM_LUN_ACTIVE,
	TGTADM_DRIVER_ACTIVE,
	TGTADM_UNSUPPORTED_OPERATION,
	TGTADM_UNKNOWN_PARAM,

	TGTADM_PREVENT_REMOVAL,
};

typedef enum tgtadm_errno tgtadm_err;

#endif
